/**
 * Clase que represanta a un polinomio, e implementa las interfaz
 * java.lang.Comparable.
 * Para ello nos apoyamos en un array de enteros, en donde irán los respectivos
 * coeficientes.<br><br>
 * <u>NOTA</u>: En la documentación nos referiremos a la clase o un objeto
 * Polinomio nombrándolo en mayúsuculas. Y si nos referiremos a un polinomio
 * desde el punto de vista matemático, lo haremos en minúsculas.
 * @see java.lang.Comparable
 * @see java.lang.Cloneable
 * @version	1.2.1
 * @author	Rayco Alonso de la Rosa (alu2201)
 */
public class Polinomio implements Comparable, Cloneable {
	/**
	 * Constante entera que da el tamaño máximo que puede tener un polinomio.
	 */
	protected static final int max_len = 25;

	/**
	 * Array de enteros con los coeficientes del polinomio.
	 */
	protected int[] poly;

	/**
	 * Entero que da el tamaño actual del array de coeficientes.<br><br>
	 * Nótese que 'len' vale siempre uno más que el grado del polinomio.
	 */
	protected int len;

	/**
	 * Constructor vacío (grado 0).
	 */
	public Polinomio() {
		poly = new int[max_len];
		len = 1;
	}

	/**
	 * Contructor a partir de una constante (grado 0).
	 * @param	i	Entero que será la constante o coeficiente de grado 0
	 */
	public Polinomio(int i) {
		poly = new int[max_len];
		poly[0] = i;
		len = 1;
	}

	/**
	 * Constructor a partir de un array.
	 * @param	ai	Array de enteros que representa a los coeficientes
	 */
	public Polinomio(int[] ai) {
		poly = new int[max_len];
		for (int i = ai.length-1; i >= 0; i--) {
			if (ai[i] != 0) {
				len = i + 1;
				break;
			}
			if (i == 0) {
				len = 1;
				break;
			}
		}
		for (int i = 0; i < len; i++)
			poly[i] = ai[i];
	}

	/**
	 * Devuelve la longitud del array de coeficientes del polinomio ('len', que
	 * es privado).
	 * @return	La longitud del array de coeficientes del polinomio
	 */
	public int getLen() {
		return len;
	}

	/**
	 * Devuelve el coeficiente enésimo del array de coeficientes ('poly', que
	 * es privado).
	 * @param	n	Entero con la posición enésima del array 'poly' a devolver
	 * @return	El coeficiente dado por poly[n], o '0' si estamos fuera de rango
	 */
	public int getPoly_n(int n) {
		if ( (n >= 0) && (n < max_len) )
			return poly[n];
		else
			return 0;
	}

	/**
	 * Evalua el Polinomio en un punto 'x' y devuelve el valor calculado.
	 * @param	x	Double con el valor que sustituirá a la X en el polinomio
	 * @return	El valor calculado de sustituir la X en todo el polinomio
	 */
	public double evalue(double x) {
		double sum = 0.0;
		for (int i = 0; i < len; i++)
			sum += poly[i] * Math.pow(x, (double) i);
		return sum;
	}

	/**
	 * Evalua el Polinomio en un punto 'x' y devuelve el valor calculado.
	 * @param	x	Float con el valor que sustituirá a la X en el polinomio
	 * @return	El valor calculado de sustituir la X en todo el polinomio
	 */
	public float evalue(float x) {
		float sum = 0.0f;
		for (int i = 0; i < len; i++)
			sum += poly[i] * Math.pow( (double)x, (double)i );
		return sum;
	}

	/**
	* Suma el Polinomio actual con otro Polinomio 'B' y lo devuelve.<br><br>
	* La suma se hace en un array para no modificar el Polinomio actual.
	* @param	B	Polinomio que se sumará con el actual
	* @return	Nuevo Polinomio con la suma del Polinomio actual con 'B'
	*/
	public Polinomio add(Polinomio B) {
		int max = (len >= B.len) ? len : B.len;
		int[] ai = new int[max];
		for (int i = 0; i < max; i++)
			ai[i] = poly[i] + B.poly[i];
		return (new Polinomio(ai));
	}

	/**
	 * Resta el Polinomio actual con otro Polinomio 'B' y lo devuelve.<br><br>
	 * La resta se hace en un array para no modificar el Polinomio actual.
	 * @param	B	Polinomio que se restará con el actual
	 * @return	Nuevo Polinomio con la resta del Polinomio actual con 'B'
	 */
	public Polinomio subtract(Polinomio B) {
		int max = (len >= B.len) ? len : B.len;
		int[] ai = new int[max];
		for (int i = 0; i < max; i++)
			ai[i] = poly[i] - B.poly[i];
		return (new Polinomio(ai));
	}

	/**
	 * Multiplica el Polinomio actual con otro Polinomio 'B' y lo devuelve.<br>
	 * <br>La multiplicación se hace en un array para no modificar el Polinomio
	 * actual.
	 * @param	B	Polinomio que se multiplicará con el actual
	 * @return	Nuevo Polinomio con la multiplicación del actual con 'B'
	 */
	public Polinomio multiply(Polinomio B) {
		int max = len + B.len - 1;
		int[] ai = new int[max];
		for (int i = 0; i < len; i++)
			for (int j = 0; j < B.len; j++)
				ai[i+j] += poly[i] * B.poly[j];
		return (new Polinomio(ai));
	}

	/**
	 * Compara si el Polinomio actual es igual al Polinomio 'B' y duvuelve un
	 * booleano.
	 * @param	B	Polinomio que se comparará con el actual
	 * @return	Booleano que nos dirá si los dos polinomios son iguales o no
	 */
	public boolean equals(Polinomio B) {
		if (len != B.len)
			return false;
		boolean flag = true;
		for (int i = 0; (i < len) && flag; i++) {
			if (poly[i] != B.poly[i])
				flag = false;
		}
		return flag;
	}

	/**
	 * Devuelve el polinomio en un String tal que así: 3X^2 -X +2
	 * @return	String que representa al polinomio de forma amigable
	 */
	public String toString() {
		String out = new String();
		boolean zero_flag = true;
		for (int i = len-1; i >= 0; i--) {
			if (poly[i] != 0) {
				zero_flag = false;
				if (poly[i] > 0)
					out += "+";
				switch (i) {
					case 0:
						out += poly[i];
						break;
					case 1:
						if ( (poly[i] != 1) && (poly[i] != -1) )
							out += poly[i];
						if (poly[i] == -1)
							out += "-";
						out += "X ";
						break;
					default:
						if ( (poly[i] != 1) && (poly[i] != -1) )
							out += poly[i];
						if (poly[i] == -1)
							out += "-";
						out += "X^" + i + " ";
						break;
				}
			}
		}
		if (zero_flag)
			out = "0";
		return out;
	}

	/**
	 * Devuelve el polinomio en un String sólo con los coeficientes separados
	 * por un espacio.
	 * @return	String que representa al polinomio (sólo coeficientes)
	 */
	public String toString2() {
		String out = new String();

		for (int i = len-1; i >= 0; i--) {
			if (poly[i] >= 0)
				out += " ";
			out += poly[i] + " ";
		}
		return out;
	}

	/**
	 * Método clone de Object, en donde la interfaz Cloneable lo permite hacer
	 * públicar, invalidando el anterior.
	 * Devuelve una copia exacta del Polinomio actual.
	 * @see java.lang.Cloneable
	 * @see java.lang.Object#clone
	 * @return	Objeto (Polinomio) que se devolverá ya copiado
	 */
	public Object clone() {
		Object o = null;
		try {
			o = super.clone();
		} catch(CloneNotSupportedException ex) {
			System.out.println("Error al clonar");
		}
		return o;
	}

	/**
	 * Método que se implementa de la interfaz java.lang.Comparable, el cual
	 * devuelve 0 si el Polinomio actual es igual que se pasa por parámetro, -1
	 * si es menor y 1 si es mayor.
	 * @see java.lang.Comparable
	 * @param	o	Objeto (Polinomio) que se comparará con el actual
	 * @return	Entero que nos dirá si es igual (0), mayor (1), o menor (-1)
	 */
	public int compareTo(Object o) {
		Polinomio po = (Polinomio)o;

		if (len > po.len)
			return 1;
		else if (len < po.len)
			return -1;
		else {
			for (int i = len-1; i >= 0; i--) {
				if (poly[i] > po.poly[i])
					return 1;
				else if (poly[i] < po.poly[i])
					return -1;
			}
			return 0;
		}
	}
}